Date: 2022-10-01T06:40:47+00:00

Actual source code: ex39.c
slepc-3.18.0 2022-10-01
Report Typos and Errors

      1: /*
      2:    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      3:    SLEPc - Scalable Library for Eigenvalue Problem Computations
      4:    Copyright (c) 2002-, Universitat Politecnica de Valencia, Spain

      6:    This file is part of SLEPc.
      7:    SLEPc is distributed under a 2-clause BSD license (see LICENSE).
      8:    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      9: */
     10: /*
     11:    This example illustrates the use of Phi functions in exponential integrators.
     12:    In particular, it implements the Norsett-Euler scheme of stiff order 1.

     14:    The problem is the 1-D heat equation with source term

     16:              y_t = y_xx + 1/(1+u^2) + psi

     18:    where psi is chosen so that the exact solution is yex = x*(1-x)*exp(tend).
     19:    The space domain is [0,1] and the time interval is [0,tend].

     21:        [1] M. Hochbruck and A. Ostermann, "Explicit exponential Runge-Kutta
     22:            methods for semilinear parabolic problems", SIAM J. Numer. Anal. 43(3),
     23:            1069-1090, 2005.
     24: */

     26: static char help[] = "Exponential integrator for the heat equation with source term.\n\n"
     27:   "The command line options are:\n"
     28:   "  -n <idim>, where <idim> = dimension of the spatial discretization.\n"
     29:   "  -tend <rval>, where <rval> = real value that corresponding to the final time.\n"
     30:   "  -deltat <rval>, where <rval> = real value for the time increment.\n"
     31:   "  -combine <bool>, to represent the phi function with FNCOMBINE instead of FNPHI.\n\n";

     33: #include <slepcmfn.h>

     35: /*
     36:    BuildFNPhi: builds an FNCOMBINE object representing the phi_1 function

     38:         f(x) = (exp(x)-1)/x

     40:    with the following tree:

     42:             f(x)                  f(x)              (combined by division)
     43:            /    \                 p(x) = x          (polynomial)
     44:         a(x)    p(x)              a(x)              (combined by addition)
     45:        /    \                     e(x) = exp(x)     (exponential)
     46:      e(x)   c(x)                  c(x) = -1         (constant)
     47: */
     48: PetscErrorCode BuildFNPhi(FN fphi)
     49: {
     50:   FN             fexp,faux,fconst,fpol;
     51:   PetscScalar    coeffs[2];

     54:   FNCreate(PETSC_COMM_WORLD,&fexp);
     55:   FNCreate(PETSC_COMM_WORLD,&fconst);
     56:   FNCreate(PETSC_COMM_WORLD,&faux);
     57:   FNCreate(PETSC_COMM_WORLD,&fpol);

     59:   FNSetType(fexp,FNEXP);

     61:   FNSetType(fconst,FNRATIONAL);
     62:   coeffs[0] = -1.0;
     63:   FNRationalSetNumerator(fconst,1,coeffs);

     65:   FNSetType(faux,FNCOMBINE);
     66:   FNCombineSetChildren(faux,FN_COMBINE_ADD,fexp,fconst);

     68:   FNSetType(fpol,FNRATIONAL);
     69:   coeffs[0] = 1.0; coeffs[1] = 0.0;
     70:   FNRationalSetNumerator(fpol,2,coeffs);

     72:   FNSetType(fphi,FNCOMBINE);
     73:   FNCombineSetChildren(fphi,FN_COMBINE_DIVIDE,faux,fpol);

     75:   FNDestroy(&faux);
     76:   FNDestroy(&fpol);
     77:   FNDestroy(&fconst);
     78:   FNDestroy(&fexp);
     79:   return 0;
     80: }

     82: int main(int argc,char **argv)
     83: {
     84:   Mat               L;
     85:   Vec               u,w,z,yex;
     86:   MFN               mfnexp,mfnphi;
     87:   FN                fexp,fphi;
     88:   PetscBool         combine=PETSC_FALSE;
     89:   PetscInt          i,k,Istart,Iend,n=199,steps;
     90:   PetscReal         t,tend=1.0,deltat=0.01,nrmd,nrmu,x,h;
     91:   const PetscReal   half=0.5;
     92:   PetscScalar       value,c,uval,*warray;
     93:   const PetscScalar *uarray;

     96:   SlepcInitialize(&argc,&argv,(char*)0,help);

     98:   PetscOptionsGetInt(NULL,NULL,"-n",&n,NULL);
     99:   PetscOptionsGetReal(NULL,NULL,"-tend",&tend,NULL);
    100:   PetscOptionsGetReal(NULL,NULL,"-deltat",&deltat,NULL);
    101:   PetscOptionsGetBool(NULL,NULL,"-combine",&combine,NULL);
    102:   h = 1.0/(n+1.0);
    103:   c = (n+1)*(n+1);

    105:   steps = (PetscInt)(tend/deltat);
    107:   PetscPrintf(PETSC_COMM_WORLD,"\nHeat equation via phi functions, n=%" PetscInt_FMT ", tend=%g, deltat=%g%s\n\n",n,(double)tend,(double)deltat,combine?" (combine)":"");

    109:   /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    110:                  Build the 1-D Laplacian and various vectors
    111:      - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    112:   MatCreate(PETSC_COMM_WORLD,&L);
    113:   MatSetSizes(L,PETSC_DECIDE,PETSC_DECIDE,n,n);
    114:   MatSetFromOptions(L);
    115:   MatSetUp(L);
    116:   MatGetOwnershipRange(L,&Istart,&Iend);
    117:   for (i=Istart;i<Iend;i++) {
    118:     if (i>0) MatSetValue(L,i,i-1,c,INSERT_VALUES);
    119:     if (i<n-1) MatSetValue(L,i,i+1,c,INSERT_VALUES);
    120:     MatSetValue(L,i,i,-2.0*c,INSERT_VALUES);
    121:   }
    122:   MatAssemblyBegin(L,MAT_FINAL_ASSEMBLY);
    123:   MatAssemblyEnd(L,MAT_FINAL_ASSEMBLY);
    124:   MatCreateVecs(L,NULL,&u);
    125:   VecDuplicate(u,&yex);
    126:   VecDuplicate(u,&w);
    127:   VecDuplicate(u,&z);

    129:   /*
    130:      Compute various vectors:
    131:      - the exact solution yex = x*(1-x)*exp(tend)
    132:      - the initial condition u = abs(x-0.5)-0.5
    133:   */
    134:   for (i=Istart;i<Iend;i++) {
    135:     x = (i+1)*h;
    136:     value = x*(1.0-x)*PetscExpReal(tend);
    137:     VecSetValue(yex,i,value,INSERT_VALUES);
    138:     value = PetscAbsReal(x-half)-half;
    139:     VecSetValue(u,i,value,INSERT_VALUES);
    140:   }
    141:   VecAssemblyBegin(yex);
    142:   VecAssemblyBegin(u);
    143:   VecAssemblyEnd(yex);
    144:   VecAssemblyEnd(u);
    145:   VecViewFromOptions(yex,NULL,"-exact_sol");
    146:   VecViewFromOptions(u,NULL,"-initial_cond");

    148:   /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    149:               Create two MFN solvers, for exp() and phi_1()
    150:      - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    151:   MFNCreate(PETSC_COMM_WORLD,&mfnexp);
    152:   MFNSetOperator(mfnexp,L);
    153:   MFNGetFN(mfnexp,&fexp);
    154:   FNSetType(fexp,FNEXP);
    155:   FNSetScale(fexp,deltat,1.0);
    156:   MFNSetErrorIfNotConverged(mfnexp,PETSC_TRUE);
    157:   MFNSetFromOptions(mfnexp);

    159:   MFNCreate(PETSC_COMM_WORLD,&mfnphi);
    160:   MFNSetOperator(mfnphi,L);
    161:   MFNGetFN(mfnphi,&fphi);
    162:   if (combine) BuildFNPhi(fphi);
    163:   else {
    164:     FNSetType(fphi,FNPHI);
    165:     FNPhiSetIndex(fphi,1);
    166:   }
    167:   FNSetScale(fphi,deltat,1.0);
    168:   MFNSetErrorIfNotConverged(mfnphi,PETSC_TRUE);
    169:   MFNSetFromOptions(mfnphi);

    171:   /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    172:              Solve the problem with the Norsett-Euler scheme
    173:      - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    174:   t = 0.0;
    175:   for (k=0;k<steps;k++) {

    177:     /* evaluate nonlinear part */
    178:     VecGetArrayRead(u,&uarray);
    179:     VecGetArray(w,&warray);
    180:     for (i=Istart;i<Iend;i++) {
    181:       x = (i+1)*h;
    182:       uval = uarray[i-Istart];
    183:       value = x*(1.0-x)*PetscExpReal(t);
    184:       value = value + 2.0*PetscExpReal(t) - 1.0/(1.0+value*value);
    185:       value = value + 1.0/(1.0+uval*uval);
    186:       warray[i-Istart] = deltat*value;
    187:     }
    188:     VecRestoreArrayRead(u,&uarray);
    189:     VecRestoreArray(w,&warray);
    190:     MFNSolve(mfnphi,w,z);

    192:     /* evaluate linear part */
    193:     MFNSolve(mfnexp,u,u);
    194:     VecAXPY(u,1.0,z);
    195:     t = t + deltat;

    197:   }
    198:   VecViewFromOptions(u,NULL,"-computed_sol");

    200:   /*
    201:      Compare with exact solution and show error norm
    202:   */
    203:   VecCopy(u,z);
    204:   VecAXPY(z,-1.0,yex);
    205:   VecNorm(z,NORM_2,&nrmd);
    206:   VecNorm(u,NORM_2,&nrmu);
    207:   PetscPrintf(PETSC_COMM_WORLD," The relative error at t=%g is %.4f\n\n",(double)t,(double)(nrmd/nrmu));

    209:   /*
    210:      Free work space
    211:   */
    212:   MFNDestroy(&mfnexp);
    213:   MFNDestroy(&mfnphi);
    214:   MatDestroy(&L);
    215:   VecDestroy(&u);
    216:   VecDestroy(&yex);
    217:   VecDestroy(&w);
    218:   VecDestroy(&z);
    219:   SlepcFinalize();
    220:   return 0;
    221: }

    223: /*TEST

    225:    test:
    226:       suffix: 1
    227:       args: -n 127 -tend 0.125 -mfn_tol 1e-3 -deltat 0.025
    228:       timeoutfactor: 2

    230:    test:
    231:       suffix: 2
    232:       args: -n 127 -tend 0.125 -mfn_tol 1e-3 -deltat 0.025 -combine
    233:       filter: sed -e "s/ (combine)//"
    234:       requires: !single
    235:       output_file: output/ex39_1.out
    236:       timeoutfactor: 2

    238: TEST*/